# main entry point for CLI applications.

import logging
import os
import argparse
from groovylight.config import Config
from groovylight.platforms.cxxrtl_sim import emit_cxxrtl

logger = logging.getLogger(__loader__.name)


def setup_logger(args):
    root_logger = logging.getLogger()
    handler = logging.StreamHandler()

    formatter = logging.Formatter(
        style=r"{", fmt=r"{levelname:s}: {name:s}: {message:s}"
    )

    handler.setFormatter(formatter)

    root_logger.addHandler(handler)
    if args.log_file is not None:
        hdlr = logging.FileHandler(args.log_file)
        hdlr.setFormatter(formatter)
        root_logger.addHandler(formatter)

    root_logger.setLevel(args.loglevel)


def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise NotADirectoryError(string)


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-d",
        "--debug",
        help="Print debug-level statements",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.INFO,
    )

    parser.add_argument(
        "-L",
        "--log-file",
        help="Log to file",
        type=argparse.FileType("w"),
        metavar="FILE",
    )
    parser.add_argument(
        "-D", "--dump", help="Dump verilog to folder", type=dir_path, metavar="FOLDER"
    )

    parser.add_argument(
        "config",
        help="Configuration file",
        type=argparse.FileType("rb"),
        metavar="FILE",
    )

    args = parser.parse_args()

    setup_logger(args)

    conf = Config(args.config)

    # use the config to create the module.

    if conf.conf["hardware"]["type"] == "cxxrtl":
        logger.info("Generating CXXRTL based graphical simulator.")
        emit_cxxrtl(conf)

    elif conf.conf["hardware"]["type"] == "colorlight":
        logger.debug("Generating colorlight code")
        if args.dump:
            logger.info(f"Dumping verilog to {args.dump}")


if __name__ == "__main__":
    main()
