from amaranth import Module
from amaranth.lib import wiring, data
from amaranth.lib.memory import Memory
from amaranth.sim import Simulator
import random
from random import randrange
import pytest

from groovylight.common import Rgb888Layout, Rgb666Layout
from groovylight.geom import DisplayGeometry, DisplayString, DisplayRotation, DisplayDimensions, Coord

from groovylight.hub75 import (
    DisplayClock,
    Hub75Coordinator,
    Hub75DataDriver,
    SwapBuffer,
)

random.seed()

def test_swapbuffer():
    dut = SwapBuffer(Rgb666Layout, 512)
    sim = Simulator(dut)
    sim.add_clock(1e-6)

    async def testbench(ctx):
        init_color = {"red": 0, "green": 0, "blue": 0}
        test_color = {"red": 8, "green": 8, "blue": 8}
        ctx.set(dut.selector, 0)
        ctx.set(dut.write_port.addr, 1)
        ctx.set(dut.read_port.addr, 1)
        ctx.set(dut.write_port.data, test_color)
        await ctx.tick()
        # assert that the read port addr 1 = 0
        assert ctx.get(dut.read_port.data) == init_color
        # swap buffer
        ctx.set(dut.selector, 1)
        await ctx.tick().repeat(
            2
        )  # takes two clocks after switching selector to output data.
        assert ctx.get(dut.read_port.data) == test_color

    # TODO: add more assertions/verification
    sim.add_testbench(testbench)
    with sim.write_vcd("output.vcd"):
        sim.run_until(1e-6 * 1000)


def randcolor(nbits=8):

    return {
        "red": randrange(2**nbits),
        "green": randrange(2**nbits),
        "blue": randrange(2**nbits),
    }


def make_start(dut, bcm):
    async def testbench(ctx):
        # select a bit, strobe start, read values, test against known.
        ctx.set(dut.bcm_select, bcm)
        ctx.set(dut.start, 1)
        await ctx.tick()
        ctx.set(dut.start, 0)
        assert ctx.get(dut.bram.en) == 1
        await ctx.tick().until(dut.done == 1)
    return testbench

@pytest.mark.parametrize("bcm", range(8))
def test_datadriver(bcm):
    memdata = [randcolor() for x in range(256)]
    m = Module()
    m.submodules.dut = dut = Hub75DataDriver()
    m.submodules.mem = mem = Memory(shape=Rgb888Layout, depth=256, init=memdata)
    m.submodules.clocker = clocker = DisplayClock()
    m.d.comb += [
        clocker.start.eq(dut.start),
        clocker.done.eq(dut.done),
    ]

    wiring.connect(m, mem.read_port(), dut.bram)

    testbench = make_start(dut, bcm)

    async def rgbtest(ctx):
        await ctx.tick().until(dut.start == 1)
        counter = 127
        async for _, rgb0, rgb1 in ctx.posedge(clocker.display_clk).sample(
            dut.data.rgb0, dut.data.rgb1
        ):
            assert counter >= 0, "should not do more than 128 clocks"
            e0 = ctx.get(mem.data[counter << 1])
            e1 = ctx.get(mem.data[(counter << 1) + 1])
            for r, e in [(rgb0, e0), (rgb1, e1)]:
                assert r.red == (e.red >> bcm) & 1
                assert r.green == (e.green >> bcm) & 1
                assert r.blue == (e.blue >> bcm) & 1
            counter = counter - 1

    sim = Simulator(m)
    sim.add_clock(1e-6)
    sim.add_testbench(testbench)
    sim.add_testbench(rgbtest, background=True)

    with sim.write_vcd("output.vcd"):
        sim.run()


@pytest.mark.parametrize("bcm", range(6))
def test_datadriver_single(bcm):
    memdata = [[randcolor(6), randcolor(6)] for x in range(128)]
    m = Module()
    m.submodules.dut = dut = Hub75DataDriver(
        data_shape=Rgb666Layout, double_fetch=False
    )
    mshape = data.ArrayLayout(Rgb666Layout, 2)
    m.submodules.mem = mem = Memory(shape=mshape, depth=128, init=memdata)
    m.submodules.clocker = clocker = DisplayClock(double_fetch = False)
    m.d.comb += [
        clocker.start.eq(dut.start),
        clocker.done.eq(dut.done),
    ]

    wiring.connect(m, mem.read_port(), dut.bram)

    testbench = make_start(dut, bcm)

    async def rgbtest(ctx):
        await ctx.tick().repeat(2)
        counter = 127
        async for _, rgb0, rgb1 in ctx.posedge(clocker.display_clk).sample(
            dut.data.rgb0, dut.data.rgb1
        ):
            assert counter >= 0, "should not do more than 128 clocks"
            e0, e1 = ctx.get(mem.data[counter])
            for r, e in [(rgb0, e0), (rgb1, e1)]:
                assert r.red == (e.red >> bcm) & 1
                assert r.green == (e.green >> bcm) & 1
                assert r.blue == (e.blue >> bcm) & 1
            counter = counter - 1

    sim = Simulator(m)
    sim.add_clock(1e-6)
    sim.add_testbench(testbench)
    sim.add_testbench(rgbtest, background=True)

    with sim.write_vcd("output.vcd"):
        sim.run()


def test_hub75_coordinator():
    m = Module()
    geom = DisplayGeometry()
    geom.add_string(DisplayString(Coord(0,0), DisplayDimensions(128,64), DisplayRotation.R0))
    geom.add_string(DisplayString(Coord(65,0), DisplayDimensions(128,64), DisplayRotation.R0))
    m.submodules.dut = dut = Hub75Coordinator(geom)

    sim = Simulator(m)
    sim.add_clock(1e-6)

    with sim.write_vcd("output.vcd"):
        sim.run_until(1e-6 * 1000)
